@@ -1,10 +1,10 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 
-# Copyright (c) 2009-2013
+# Copyright (c) 2009-2014
 
 # Author(s):
- 
+
 #   Martin Raspaud <martin.raspaud@smhi.se>
 #   Adam Dybbroe <adam.dybbroe@smhi.se>
 #   Esben S. Nielsen <esn@dmi.dk>
@@ -92,35 +92,35 @@ class Image(object):
     for example will produce black where data is missing."None" (default) will
     produce transparency (thus adding an alpha channel) if the file format
     allows it, black otherwise.
-    
+
     The channels are considered to contain floating point values in the range
     [0.0,1.0]. In order to normalize the input data, the *color_range*
     parameter defines the original range of the data. The conversion to the
     classical [0,255] range and byte type is done automagically when saving the
     image to file.
     """
-    channels = None
-    mode = None
-    width = 0
-    height = 0
-    fill_value = None
-    palette = None
-
-    _secondary_mode = "RGB"
 
     modes = ["L", "LA", "RGB", "RGBA", "YCbCr", "YCbCrA", "P", "PA"]
-    
-    #: Shape (dimensions) of the image.
-    shape = None
-    
-    def __init__(self, channels=None, mode="L", color_range=None, 
+
+    def __init__(self, channels=None, mode="L", color_range=None,
                  fill_value=None, palette=None):
 
+        self.channels = None
+        self.mode = None
+        self.width = 0
+        self.height = 0
+        self.fill_value = None
+        self.palette = None
+        self.shape = None
+        self.info = {}
+
+        self._secondary_mode = "RGB"
+
         if(channels is not None and
            not isinstance(channels, (tuple, set, list,
                                      np.ndarray, np.ma.core.MaskedArray))):
-            raise TypeError("Channels should a tuple, set, list, numpy array, "
-                            "or masked array.")
+            raise TypeError("Image channels should a tuple, set, list, numpy "
+                            "array, or masked array.")
 
         if(isinstance(channels, (tuple, list)) and
            len(channels) != len(re.findall("[A-Z]", mode))):
@@ -137,13 +137,13 @@ class Image(object):
         if(color_range is not None and
            _is_list_of_pairs(color_range) and
            (channels is None or
-            len(color_range) != len(channels))):
+             len(color_range) != len(channels))):
             raise ValueError("Color_range length does not match number of "
                              "channels.")
-        
+
         if(color_range is not None and
            (((mode == "L" or mode == "P") and not _is_pair(color_range)) and
-            (len(color_range) != len(re.findall("[A-Z]", mode))))):
+             (len(color_range) != len(re.findall("[A-Z]", mode))))):
             raise ValueError("Color_range does not match mode")
 
         self.mode = mode
@@ -154,7 +154,7 @@ class Image(object):
             self.fill_value = [fill_value]
         else:
             self.fill_value = None
-            
+
         self.channels = []
         self.palette = palette
 
@@ -168,18 +168,18 @@ class Image(object):
                     else:
                         color_min = 0.0
                         color_max = 1.0
-                    
+
                     # Add data to image object as a channel
                     self._add_channel(chn, color_min, color_max)
 
                     self.shape = self.channels[-1].shape
                     if self.shape != self.channels[0].shape:
-                        raise ValueError("Channels must have the same shape.")
+                        raise ValueError("Image channels must have the same"
+                                         " shape.")
                 self.height = self.shape[0]
                 self.width = self.shape[1]
             else:
-                raise ValueError("Channels must all be arrays, lists or "
-                                 "tuples.")
+                raise ValueError("Image channels must all be arrays tuples.")
         elif channels is not None:
             self.height = channels.shape[0]
             self.width = channels.shape[1]
@@ -191,7 +191,7 @@ class Image(object):
             else:
                 color_min = 0.0
                 color_max = 1.0
-            
+
             # Add data to image object as a channel
             self._add_channel(channels, color_min, color_max)
 
@@ -203,20 +203,21 @@ class Image(object):
     def _add_channel(self, chn, color_min, color_max):
         """Adds a channel to the image object
         """
-        
+
         if isinstance(chn, np.ma.core.MaskedArray):
             chn_data = chn.data
             chn_mask = chn.mask
         else:
             chn_data = np.array(chn)
             chn_mask = False
-        scaled = ((chn_data - color_min) * 
+        scaled = ((chn_data - color_min) *
                   1.0 / (color_max - color_min))
         self.channels.append(np.ma.array(scaled, mask=chn_mask))
-    
-    def _finalize(self):
+
+    def _finalize(self, dtype=np.uint8):
         """Finalize the image, that is put it in RGB mode, and set the channels
-        in 8bit format ([0,255] range).
+        in unsigned 8bit format ([0,255] range) (if the *dtype* doesn't say
+        otherwise).
         """
         channels = []
         if self.mode == "P":
@@ -226,15 +227,16 @@ class Image(object):
 
         for chn in self.channels:
             if isinstance(chn, np.ma.core.MaskedArray):
-                final_data = chn.data.clip(0, 1) * 255
+                final_data = chn.data.clip(0, 1) * np.iinfo(dtype).max
             else:
-                final_data = chn.clip(0, 1) * 255
-                
+                final_data = chn.clip(0, 1) * np.iinfo(dtype).max
+
             channels.append(np.ma.array(final_data,
-                                        np.uint8,
-                                        mask = chn.mask))
+                                        dtype,
+                                        mask = np.ma.getmaskarray(chn)))
         if self.fill_value is not None:
-            fill_value = [int(col * 255) for col in self.fill_value]
+            fill_value = [int(col * np.iinfo(dtype).max)
+                          for col in self.fill_value]
         else:
             fill_value = None
         return channels, fill_value
@@ -252,7 +254,7 @@ class Image(object):
         """Display the image on screen.
         """
         self.pil_image().show()
-        
+
 
     def pil_image(self):
         """Return a PIL image from the current image.
@@ -262,7 +264,7 @@ class Image(object):
         if self.is_empty():
             return Pil.new(self.mode, (0, 0))
 
-        if(self.mode == "L"):
+        if self.mode == "L":
             if fill_value is not None:
                 img = Pil.fromarray(channels[0].filled(fill_value))
             else:
@@ -271,9 +273,9 @@ class Image(object):
                 mask = np.ma.getmaskarray(channels[0])
                 alpha = np.where(mask, alpha, 255)
                 pil_alpha = Pil.fromarray(alpha)
-                
+
                 img = Pil.merge("LA", (img, pil_alpha))
-        elif(self.mode == "LA"):
+        elif self.mode == "LA":
             if fill_value is not None:
                 img = Pil.fromarray(channels[0].filled(fill_value))
                 pil_alpha = Pil.fromarray(channels[1])
@@ -285,11 +287,11 @@ class Image(object):
                 pil_alpha = Pil.fromarray(alpha)
             img = Pil.merge("LA", (img, pil_alpha))
 
-        elif(self.mode == "RGB"):
+        elif self.mode == "RGB":
             # Mask where all channels have missing data (incomplete data will
             # be shown).
-            mask = (np.ma.getmaskarray(channels[0]) & 
-                    np.ma.getmaskarray(channels[1]) & 
+            mask = (np.ma.getmaskarray(channels[0]) &
+                    np.ma.getmaskarray(channels[1]) &
                     np.ma.getmaskarray(channels[2]))
 
             if fill_value is not None:
@@ -308,11 +310,11 @@ class Image(object):
 
                 img = Pil.merge("RGBA", (pil_r, pil_g, pil_b, pil_a))
 
-        elif(self.mode == "RGBA"):
+        elif self.mode == "RGBA":
             # Mask where all channels have missing data (incomplete data will
             # be shown).
-            mask = (np.ma.getmaskarray(channels[0]) & 
-                    np.ma.getmaskarray(channels[1]) & 
+            mask = (np.ma.getmaskarray(channels[0]) &
+                    np.ma.getmaskarray(channels[1]) &
                     np.ma.getmaskarray(channels[2]) &
                     np.ma.getmaskarray(channels[3]))
 
@@ -351,17 +353,17 @@ class Image(object):
         """
         # PIL does not support compression option.
         del compression
-        
+
         if self.is_empty():
             raise IOError("Cannot save an empty image")
-        
+
         ensure_dir(filename)
 
         fformat = fformat or os.path.splitext(filename)[1][1:4]
         fformat = check_image_format(fformat)
 
         self.pil_image().save(filename, fformat)
-                    
+
     def putalpha(self, alpha):
         """Adds an *alpha* channel to the current image, or replaces it with
         *alpha* if it already exists.
@@ -371,12 +373,12 @@ class Image(object):
                 self.shape[0] == 0) and
            alpha.shape != self.shape):
             raise ValueError("Alpha channel shape should match image shape")
-        
-        if(not self.mode.endswith("A")):
+
+        if not self.mode.endswith("A"):
             self.convert(self.mode+"A")
         if not self.is_empty():
             self.channels[-1] = alpha
-        
+
     def _rgb2ycbcr(self, mode):
         """Convert the image from RGB mode to YCbCr."""
 
@@ -394,10 +396,10 @@ class Image(object):
                                              self.fill_value[2])
 
         self.mode = mode
-        
+
     def _ycbcr2rgb(self, mode):
         """Convert the image from YCbCr mode to RGB.
-        """ 
+        """
 
         self._check_modes(("YCbCr", "YCbCrA"))
 
@@ -426,12 +428,12 @@ class Image(object):
             chans = self.channels
             alpha = None
             self._secondary_mode = self.mode
-        
+
         palette = []
         selfmask = reduce(np.ma.mask_or, [chn.mask for chn in chans])
         new_chn = np.ma.zeros(self.shape, dtype = int)
         color_nb = 0
-        
+
         for i in range(self.height):
             for j in range(self.width):
                 current_col = tuple([chn[i, j] for chn in chans])
@@ -443,7 +445,7 @@ class Image(object):
                     idx = color_nb
                     palette.append(current_col)
                     color_nb = color_nb + 1
-                    
+
                 new_chn[i, j] = idx
 
         if self.fill_value is not None:
@@ -513,7 +515,7 @@ class Image(object):
             self.mode = self.mode + "A"
 
         self.convert(mode)
-                                        
+
 
     def _check_modes(self, modes):
         """Check that the image is in on of the given *modes*, raise an
@@ -523,7 +525,7 @@ class Image(object):
             modes = [modes]
         if self.mode not in modes:
             raise ValueError("Image not in suitable mode: %s"%modes)
-            
+
 
     def _l2rgb(self, mode):
         """Convert from L (black and white) to RGB.
@@ -545,11 +547,11 @@ class Image(object):
 
         kb_ = 0.114
         kr_ = 0.299
-        
+
         r__ = self.channels[0]
         g__ = self.channels[1]
         b__ = self.channels[2]
-        
+
         y__ = kr_ * r__ + (1 - kr_ - kb_) * g__ + kb_ * b__
 
         if self.fill_value is not None:
@@ -559,10 +561,10 @@ class Image(object):
                                self.fill_value[3:])
 
         self.channels = [y__] + self.channels[3:]
-        
+
         self.mode = mode
 
-        
+
     def _ycbcr2l(self, mode):
         """Convert from YCbCr to L.
         """
@@ -577,7 +579,7 @@ class Image(object):
         """Convert from L to YCbCr.
         """
         self._check_modes(("L", "LA"))
-        
+
         luma = self.channels[0]
         zeros = np.ma.zeros(luma.shape)
         zeros.mask = luma.mask
@@ -590,7 +592,7 @@ class Image(object):
 
         self.mode = mode
 
-        
+
 
     def convert(self, mode):
         """Convert the current image to the given *mode*. See :class:`Image`
@@ -606,24 +608,24 @@ class Image(object):
         if self.is_empty():
             self.mode = mode
             return
-        
-        if(mode == self.mode + "A"):
+
+        if mode == self.mode + "A":
             self.channels.append(np.ma.ones(self.channels[0].shape))
             if self.fill_value is not None:
                 self.fill_value += [1]
             self.mode = mode
 
-        elif(mode + "A" == self.mode):
+        elif mode + "A" == self.mode:
             self.channels = self.channels[:-1]
             if self.fill_value is not None:
                 self.fill_value = self.fill_value[:-1]
             self.mode = mode
 
-        elif(mode.endswith("A") and not self.mode.endswith("A")):
+        elif mode.endswith("A") and not self.mode.endswith("A"):
             self.convert(self.mode + "A")
             self.convert(mode)
 
-        elif(self.mode.endswith("A") and not mode.endswith("A")):
+        elif self.mode.endswith("A") and not mode.endswith("A"):
             self.convert(self.mode[:-1])
             self.convert(mode)
 
@@ -658,13 +660,13 @@ class Image(object):
             except KeyError:
                 raise ValueError("Conversion from %s to %s not implemented !"
                                  %(self.mode,mode))
-            
+
     def clip(self, channels = True):
         """Limit the values of the array to the default [0,1] range. *channels*
         says which channels should be clipped."""
-        if not (isinstance(channels, (tuple, list))):
+        if not isinstance(channels, (tuple, list)):
             channels = [channels]*len(self.channels)
-            
+
         for i in range(len(self.channels)):
             if channels[i]:
                 self.channels[i] = np.ma.clip(self.channels[i], 0.0, 1.0)
@@ -691,14 +693,14 @@ class Image(object):
             factor[1] = shape[1] * 1.0 / self.width
         else:
             factor[1] = self.width * 1.0 / shape[1]
-        
+
         if(int(factor[0]) != factor[0] or
            int(factor[1]) != factor[1]):
             raise ValueError("Resize not of integer factor!")
 
         factor[0] = int(factor[0])
         factor[1] = int(factor[1])
-        
+
         i = 0
         for chn in self.channels:
             if zoom[0]:
@@ -715,7 +717,7 @@ class Image(object):
                                        [idx * factor[1]
                                         for idx in range(self.width /
                                                          factor[1])]]
-                
+
             i = i + 1
 
         self.height = self.channels[0].shape[0]
@@ -729,17 +731,17 @@ class Image(object):
         """
         if self.is_empty():
             return
-        
-        if (luminance.shape != self.channels[0].shape):
+
+        if luminance.shape != self.channels[0].shape:
             if ((luminance.shape[0] * 1.0 / luminance.shape[1]) ==
-                (self.channels[0].shape[0] * 1.0 / self.channels[0].shape[1])):
+                 (self.channels[0].shape[0] * 1.0 / self.channels[0].shape[1])):
                 if luminance.shape[0] > self.channels[0].shape[0]:
                     self.resize(luminance.shape)
                 else:
                     raise NameError("Luminance smaller than the image !")
             else:
                 raise NameError("Not the good shape !")
-        
+
         mode = self.mode
         if mode.endswith("A"):
             self.convert("YCbCrA")
@@ -749,8 +751,8 @@ class Image(object):
             self.convert("YCbCr")
             self.channels[0] = luminance
             self.convert(mode)
-            
-    def enhance(self, inverse = False, gamma = 1.0, stretch = "no"):
+
+    def enhance(self, inverse=False, gamma=1.0, stretch="no"):
         """Image enhancement function. It applies **in this order** inversion,
         gamma correction, and stretching to the current image, with parameters
         *inverse* (see :meth:`Image.invert`), *gamma* (see
@@ -760,7 +762,7 @@ class Image(object):
         self.gamma(gamma)
         self.stretch(stretch)
 
-    def gamma(self, gamma = 1.0):
+    def gamma(self, gamma=1.0):
         """Apply gamma correction to the channels of the image. If *gamma* is a
         tuple, then it should have as many elements as the channels of the
         image, and the gamma correction is applied elementwise. If *gamma* is a
@@ -777,18 +779,18 @@ class Image(object):
         if(isinstance(gamma, (list, tuple, set)) and
            len(gamma) != len(self.channels)):
             raise ValueError("Number of channels and gamma components differ.")
-        
+
         if gamma < 0:
             raise ValueError("Gamma correction must be a positive number.")
-        
+
         if gamma == 1.0:
             return
-        if (isinstance(gamma, (tuple, list))):
+        if isinstance(gamma, (tuple, list)):
             gamma_list = list(gamma)
         else:
             gamma_list = [gamma] * len(self.channels)
         for i in range(len(self.channels)):
-            if(isinstance(self.channels[i], np.ma.core.MaskedArray)):
+            if isinstance(self.channels[i], np.ma.core.MaskedArray):
                 if ne:
                     self.channels[i] = np.ma.array(
                         ne.evaluate("data ** (1.0 / gamma)",
@@ -806,8 +808,8 @@ class Image(object):
                                             self.channels[i] **
                                             (1.0 / gamma_list[i]),
                                             self.channels[i])
-        
-    def stretch(self, stretch="no", **kwarg):
+
+    def stretch(self, stretch="crude", **kwarg):
         """Apply stretching to the current image. The value of *stretch* sets
         the type of stretching applied. The values "histogram", "linear",
         "crude" (or "crude-stretch") perform respectively histogram
@@ -818,34 +820,40 @@ class Image(object):
         with the values as cutoff. These values should be normalized in the
         range [0.0,1.0].
         """
+
+        ch_len = len(self.channels)
+        if self.mode.endswith("A"):
+            ch_len -= 1
+
+
         if((isinstance(stretch, tuple) or
-            isinstance(stretch,list))):
+             isinstance(stretch, list))):
             if len(stretch) == 2:
-                for i in range(len(self.channels)):
-                    self.stretch_linear(i, cutoffs = stretch, **kwarg)
+                for i in range(ch_len):
+                    self.stretch_linear(i, cutoffs=stretch, **kwarg)
             else:
                 raise ValueError("Stretch tuple must have exactly two elements")
         elif stretch == "linear":
-            for i in range(len(self.channels)):
+            for i in range(ch_len):
                 self.stretch_linear(i, **kwarg)
         elif stretch == "histogram":
-            for i in range(len(self.channels)):
+            for i in range(ch_len):
                 self.stretch_hist_equalize(i, **kwarg)
-        elif(stretch in ["crude", "crude-stretch"]):
-            for i in range(len(self.channels)):
+        elif stretch in ["crude", "crude-stretch"]:
+            for i in range(ch_len):
                 self.crude_stretch(i, **kwarg)
-        elif(stretch in ["log", "logarithmic"]):
-            for i in range(len(self.channels)):
+        elif stretch in ["log", "logarithmic"]:
+            for i in range(ch_len):
                 self.stretch_logarithmic(i, **kwarg)
-        elif(stretch == "no"):
+        elif stretch == "no":
             return
         elif isinstance(stretch, str):
             raise ValueError("Stretching method %s not recognized."%stretch)
         else:
             raise TypeError("Stretch parameter must be a string or a tuple.")
 
-            
-    def invert(self, invert = True):
+
+    def invert(self, invert=True):
         """Inverts all the channels of a image according to *invert*. If invert
         is a tuple or a list, elementwise invertion is performed, otherwise all
         channels are inverted if *invert* is true (default).
@@ -853,20 +861,19 @@ class Image(object):
         if(isinstance(invert, (tuple, list, set)) and
            len(self.channels) != len(invert)):
             raise ValueError("Number of channels and invert components differ.")
-        
+
         if isinstance(invert, (tuple, list, set)):
             i = 0
             for chn in self.channels:
-                if(invert[i]):
+                if invert[i]:
                     self.channels[i] = 1.0 - chn
                 i = i + 1
-        elif(invert):
+        elif invert:
             i = 0
             for chn in self.channels:
                 self.channels[i] = 1.0 - chn
                 i = i + 1
-         
-   
+
     def stretch_hist_equalize(self, ch_nb):
         """Stretch the current image's colors by performing histogram
         equalization on channel *ch_nb*.
@@ -885,7 +892,7 @@ class Image(object):
         carr = arr.compressed()
 
         cdf = np.arange(0.0, 1.0, 1/nwidth)
-        logger.debug("Make histogram bins having equal amount of data, " + 
+        logger.debug("Make histogram bins having equal amount of data, " +
                   "using numpy percentile function:")
         bins = np.percentile(carr, list(cdf * 100))
 
@@ -903,7 +910,7 @@ class Image(object):
         logger.debug("Perform a logarithmic contrast stretch.")
         if ((self.channels[ch_nb].size ==
              np.ma.count_masked(self.channels[ch_nb])) or
-            (self.channels[ch_nb].min() == self.channels[ch_nb].max())):
+              (self.channels[ch_nb].min() == self.channels[ch_nb].max())):
             logger.warning("Nothing to stretch !")
             return
 
@@ -928,38 +935,36 @@ class Image(object):
            self.channels[ch_nb].min() == self.channels[ch_nb].max()):
             logger.warning("Nothing to stretch !")
             return
-        
+
         arr = self.channels[ch_nb]
         carr = arr.compressed()
 
         logger.debug("Calculate the histogram percentiles: ")
-        logger.debug("Left and right percentiles: " + 
+        logger.debug("Left and right percentiles: " +
                   str(cutoffs[0]*100) + " " + str(cutoffs[1]*100))
 
-        #left, right = (np.percentile(carr, cutoffs[0]*100),
-        #               np.percentile(carr, 100. - cutoffs[1]*100))
-        left, right = np.percentile(carr, [cutoffs[0]*100, 
+        left, right = np.percentile(carr, [cutoffs[0]*100,
                                            100. - cutoffs[1]*100])
 
         delta_x = (right - left)
-        logger.debug("Interval: left=%f, right=%f width=%f"
-                  % (left, right, delta_x))
+        logger.debug("Interval: left=%f, right=%f width=%f",
+                   left, right, delta_x)
 
         if delta_x > 0.0:
-            self.channels[ch_nb] = np.ma.array((arr - left) / delta_x, 
+            self.channels[ch_nb] = np.ma.array((arr - left) / delta_x,
                                                mask = arr.mask)
         else:
             self.channels[ch_nb] = np.ma.zeros(arr.shape)
             logger.warning("Unable to make a contrast stretch!")
 
 
-    def crude_stretch(self, ch_nb, min_stretch = None, max_stretch = None):
+    def crude_stretch(self, ch_nb, min_stretch=None, max_stretch=None):
         """Perform simple linear stretching (without any cutoff) on the channel
         *ch_nb* of the current image and normalize to the [0,1] range."""
 
-        if(min_stretch is None):
+        if min_stretch is None:
             min_stretch = self.channels[ch_nb].min()
-        if(max_stretch is None):
+        if max_stretch is None:
             max_stretch = self.channels[ch_nb].max()
 
         if((not self.channels[ch_nb].mask.all()) and
@@ -967,7 +972,7 @@ class Image(object):
             stretched = self.channels[ch_nb].data.astype(np.float)
             stretched -= min_stretch
             stretched /= max_stretch - min_stretch
-            self.channels[ch_nb] = np.ma.array(stretched, 
+            self.channels[ch_nb] = np.ma.array(stretched,
                                                mask=self.channels[ch_nb].mask,
                                                copy=False)
         else:
@@ -979,8 +984,8 @@ class Image(object):
         """
         if self.is_empty():
             raise ValueError("Cannot merge an empty image.")
-        
-        if(self.mode != img.mode):
+
+        if self.mode != img.mode:
             raise ValueError("Cannot merge image of different modes.")
 
         selfmask = reduce(np.ma.mask_or, [chn.mask for chn in self.channels])
@@ -996,7 +1001,7 @@ class Image(object):
         """Colorize the current image using
         *colormap*. Works only on"L" or "LA" images.
         """
-        
+
         if self.mode not in ("L", "LA"):
             raise ValueError("Image should be grayscale to colorize")
         if self.mode == "LA":
@@ -1014,7 +1019,7 @@ class Image(object):
         """Palettize the current image using
         *colormap*. Works only on"L" or "LA" images.
         """
-        
+
         if self.mode not in ("L", "LA"):
             raise ValueError("Image should be grayscale to colorize")
         self.channels[0], self.palette = colormap.palettize(self.channels[0])
@@ -1037,11 +1042,11 @@ class Image(object):
                                (1 - src.channels[3])) / outa
             dst.channels[i][outa == 0] = 0
         dst.channels[3] = outa
-        
+
 def _areinstances(the_list, types):
     """Check if all the elements of the list are of given type.
     """
-    return all([isinstance(item, types) for item in the_list]) 
+    return all([isinstance(item, types) for item in the_list])
 
 def _is_pair(item):
     """Check if an item is a pair (tuple of size 2).
@@ -1059,7 +1064,7 @@ def _is_list_of_pairs(the_list):
 
 def ycbcr2rgb(y__, cb_, cr_):
     """Convert the three YCbCr channels to RGB channels.
-    """ 
+    """
 
     kb_ = 0.114
     kr_ = 0.299
@@ -1070,7 +1075,6 @@ def ycbcr2rgb(y__, cb_, cr_):
 
     return r__, g__, b__
 
-        
 def rgb2ycbcr(r__, g__, b__):
     """Convert the three RGB channels to YCbCr."""
 
